//============================================================================/
//  FOUR_CAM.CPP                                                              /
//
//  Four-In-A-Row Game
//
//	This C++ source code for an Excalibur BBS Plug-In is intended only as
//	a sample of using the Alchemist's Workbench API and creating a plug-in
//	for Camelot.  Excalibur Communications makes no warranty of any kind,
//  expressed or implied, including without limitation, any warranties of
//	merchantability and/or fitness for a particular purpose.  Use this
//	source code at your own risk.
//
//====================================================================

#include "opl.h"
#include <windows.h>
#include <alchemw.h>
#include <string.h>
#include <ctype.h>
#include <dde.h>

#include "four.h"
#include "cmdfour.h"

DEFINE_LIBRARY_ENTRY;

// cheap linked list for queuing messages.
// this theoretically shouldn't be more than
// one or two in length.
struct Message {
	Message *Next;
	char Text[256];
	Message( LPCSTR text )
		: Next(0)
		{ strcpy( Text, text ); }
	};

class GameSlot : public Plugin {
private:
	BOOL SentDLL, SentHelp;
	Message *MList;
	long Opponent;			// -1 if nobody yet
	long Inviting;			// busy inviting someone
	long Inviter;			// person doing invite to this player
public:
	GameSlot(LPSLOT sloth)
		: Plugin(sloth), SentDLL(0), SentHelp(0),MList(0),
		Opponent(-1), Inviting(-1), Inviter(-1)
		{}
	BOOL NextDLL( LPSTR name, WORD limit, LPSTR dest, WORD dlim );
	WORD Type() { return AL_WIN_PLUGIN | AL_WIN_MULTI; }
	LONG OtherSizes();
	LPCSTR DLLName() { return "FOUR.DLL"; }
	WORD Receive( BYTE *data, WORD len, DWORD from );
	WORD Process();

	// processing for data from Grail
	WORD Broadcast( BYTE *data, WORD len );
	void broadAllHere();
	void broadEnter( BYTE *data, WORD len );
	void broadLeave( BYTE *data );
	void rcvInvite( DWORD from );
	void rcvReject();
	void rcvAccept();

	// processing for data from Excalibur
	virtual WORD Command( BYTE *cmd, WORD len );
	void cmdChatList( BYTE *cmd );
	void cmdTextBlock( BYTE *cmd );
	void cmdInvite( BYTE *cmd );
	void cmdResponse( BYTE *cmd );
	void cmdMove( BYTE *cmd );

	void error( int id )
		{ BYTE temp[2]; temp[0] = CMD_ERROR; temp[1]=id; AlAddBlock( temp, 2 ); }
	};

DEFINE_PLUGINS( GameSlot );
PLUGIN_UNIQUE;


BOOL GameSlot::NextDLL( LPSTR name, WORD, LPSTR dest, WORD )
	{
	OFSTRUCT ofs;

	if ( SentDLL && SentHelp )
		return FALSE;
	else if ( SentDLL ) {
		// first, see if there is a language specific version
		wsprintf( name, "FOUR%04X.HLP", AlGetUserLanguage() );
		if ( OpenFile( name, &ofs, OF_EXIST ) < 0 )
			strcpy( name, "FOUR.HLP" );
		strcpy( dest, "FOUR.HLP" );
		SentHelp = TRUE;
		return TRUE;
		}
	else {
		// first, see if there is a language specific version
		wsprintf( name, "FOUR%04X.DLL", AlGetUserLanguage() );
		if ( OpenFile( name, &ofs, OF_EXIST ) < 0 )
			strcpy( name, "FOUR.DLL" );
		strcpy( dest, "FOUR.DLL" );
		SentDLL = TRUE;
		return TRUE;
		}
	}

LONG GameSlot::OtherSizes()
	{
	// find the help file length
   char name[256];
	wsprintf( name, "FOUR%04X.HLP", AlGetUserLanguage() );
	OFSTRUCT ofs;
	HFILE in = OpenFile( name, &ofs, OF_READ );
	if ( in < 0 )
		in = OpenFile( "FOUR.HLP", &ofs, OF_READ );
	if ( in < 0 )
		return 0;
	LONG len = _llseek( in, 0, 2 );
	_lclose( in );
	return len;
	}

//============================================================================/
// These routines catch codes coming to Camelot from another Camelot.         /
//============================================================================/

WORD GameSlot::Receive( BYTE *data, WORD len, DWORD from )
	{
	switch( data[0] ) {
		case CMD_TEXTBLOCK:			// these only come from your opponent
		case CMD_MOVE:             // just pass on through
		case CMD_COLOR:
			AlAddBlock( data, len );
			break;
		case CMD_INVITE:
			rcvInvite( from );
			break;
		case CMD_ACCEPT:
			rcvAccept();
			break;
		case CMD_REJECT:
			rcvReject();
			break;
		}

	return AL_OK;
	}

void GameSlot::rcvInvite( DWORD from )
	{
	if ( Inviter !=-1 || Inviting!=-1 || Opponent != -1 ) {	// I'm busy
		AlCreateMulti( from, 1 );
		BYTE *temp = (BYTE*)AlMultiLock();
		*temp = CMD_BUSY;
		AlMultiUnlock();
		}
	else {
		struct {
			BYTE cmd;
			DWORD from;
			} data = { CMD_INVITE, from };
		AlAddBlock( &data, sizeof( data ) );
		Inviter = from;
		}
	}

void GameSlot::rcvReject()
	{
	Inviting = -1;
	BYTE temp[2];
	temp[0] = CMD_ERROR;
	temp[1] = ERR_BUSY;
	AlAddBlock( temp, 2 );
	}

void GameSlot::rcvAccept()
	{
	Opponent = Inviting;
	Inviting = -1;
	// ok, time to start the game... only one thing random is who goes first
	// (i.e., gets the red pieces)
	BYTE youfirst = (BYTE)(GetTickCount()&1);	// close enough to random

	BYTE temp[2];
	temp[0] = CMD_COLOR;
	temp[1] = ! youfirst;
	AlAddBlock( temp, 2 );

	AlCreateMulti( Opponent, 2 );
	BYTE *yourcmd = (BYTE*)AlMultiLock();
	yourcmd[0] = CMD_COLOR;
	yourcmd[1] = youfirst;
	AlMultiUnlock(); //... and we're off and running
	}

//============================================================================/
// These routines catch codes coming to Camelot from Grail (mostly having to  /
// do with people entering and leaving the game).                             /
//============================================================================/

WORD GameSlot::Broadcast( BYTE *data, WORD len )
	{
	Plugin::Broadcast( data, len );
	switch ( *data ) {
		case AL_BROAD_ALLHERE:
			broadAllHere();
			break;
		case AL_BROAD_ENTER:
			broadEnter( data, len );
			break;
		case AL_BROAD_LEAVE:
			broadLeave( data );
			break;
		}

	return AL_OK;
	}

void GameSlot::broadAllHere()
	{
	UserDictImp scan( ConnectList );
	char temp[50];
	while (scan) {
		temp[0] = CMD_WHOISON;
		*(long*)(temp+1) = scan.Current().MultiID;
		strcpy( temp+5, scan.Current().Handle );

		AlAddBlock( temp, 6+strlen( temp+5 ) );
		scan++;
		}
	temp[0] = CMD_ENDWHO;
	AlAddBlock( temp, 1 );
	}

void GameSlot::broadEnter( BYTE *data, WORD )
	{
	AlBroadEnter &entry = *(AlBroadEnter*)(data+1);
	if ( entry.MultiID == AlMultiSelf() )
		return;

	char temp[100];
	temp[0] = CMD_PERSON;
	*(long*)(temp+1) = entry.MultiID;
	strcpy( temp+5, entry.Handle );
	AlAddBlock( temp, strlen( entry.Handle ) + 6 );
	}

void GameSlot::broadLeave( BYTE *data )
	{
	char temp[100];
	temp[0] = CMD_LEAVING;
	*(long*)(temp+1) = *(long*)(data+1);
	AlAddBlock( temp, 5 );

	// check to see if our opponent left
	if ( *(long*)(data+1)==Opponent ) {
		Opponent = -1;
		}
	}


//============================================================================/
// These routines respond to commands from Excalibur to Camelot.              /
//============================================================================/

WORD GameSlot::Command( BYTE *cmd, WORD )
	{
	switch ( *cmd++ ) {
		case CMD_INVITE:
			cmdInvite( cmd );
			break;
		case CMD_ACCEPT:
		case CMD_REJECT:
			cmdResponse( cmd-1 );
			break;
		case CMD_MOVE:
			cmdMove( cmd );
			break;
		case CMD_TEXTBLOCK:
			cmdTextBlock( cmd );
			break;
		case CMD_GAMEOVER:
			Opponent = -1;
			break;
		}
	return AL_OK;
	}

// user picked someone to invite to a game
void GameSlot::cmdInvite( BYTE *cmd )
	{
	long id = *(long*)cmd;
	if ( id==AlMultiSelf() )
		error( ERR_SELF );
	else {
		if ( AlCreateMulti( id, 1 ) ) {
			BYTE *temp = (BYTE*)AlMultiLock();
			*temp = CMD_INVITE;
			AlMultiUnlock();
			Inviting = id;
			}
		}
	}

// user responded to a request to play a game
void GameSlot::cmdResponse( BYTE *cmd )
	{
	AlCreateMulti( Inviter, 1 );
	BYTE *temp = (BYTE*)AlMultiLock();
	*temp = *cmd;
	AlMultiUnlock();
	if ( *cmd==CMD_ACCEPT )
		Opponent = Inviter;
	Inviter = -1;
	}

// Received the coordinates of a move
void GameSlot::cmdMove( BYTE *cmd )
	{
	AlCreateMulti( Opponent, 3 );
	BYTE *temp = (BYTE*)AlMultiLock();
	*temp++ = CMD_MOVE;
	*temp++ = *cmd++;	// just pass values on through
	*temp++ = *cmd++;
	AlMultiUnlock();
	}

// user has a comment to the opponent
// Note:  Messages are queued and actually
// sent via 'Process'.  This is because
// only one AlCreateMulti block can be
// available at once and if you got a
// second message before the first got
// sent there could be a problem.  This
// could theoretically be the case with
// the move (above) but very unlikely for
// this game.  A more highly interactive
// game might need the moves queued like
// the text blocks are.
void GameSlot::cmdTextBlock( BYTE *cmd )
	{
	Message *m = new Message( (char*)cmd );
	Message *scan = MList;
	if ( MList ) {
		for ( scan=MList; scan->Next; scan = scan->Next )
			;
		scan->Next = m;
		}
	else
		MList = m;
	}


//============================================================================/
// Idle processing loop.  Mostly passes on any accumulated messages.          /
//============================================================================/

WORD GameSlot::Process()
	{
	if ( ! MList )
		return AL_IDLE;
	BOOL created = FALSE;
	created = AlCreateMulti( strlen( MList->Text )+2 );
	if ( ! created )
		return AL_OK;	// busy... try again in a moment

	BYTE *data = (BYTE*)AlMultiLock();
	strcpy( (char*)data+1, MList->Text );
	data[0] = CMD_TEXTBLOCK;

	AlMultiUnlock();
	Message *save = MList;
	MList = MList->Next;
	delete save;

	return AL_OK;
	}

